     /********************************************************************
      ***         Programme de test couche physique bus I2C            ***
      ***                      dsPIC30F3011                            ***
      ***                                                              ***
      ***  Auteur : CREMMEL Marcel                                     ***
      ***  Date : 21/12/2006                                           ***
      ********************************************************************/

#include <p30f3011.h>

//--------------------------Device Configuration------------------------       
_FOSC(CSW_FSCM_OFF & XT_PLL16);//Turn off clock switching and
                               //fail-safe clock monitoring and
                               //use the XT Clock x 16 as the system clock
_FWDT(WDT_OFF);                //Turn off Watchdog Timer
_FBORPOR(PBOR_ON & BORV_27 & PWRT_16 & MCLR_EN);
                               //Set Brown-out Reset voltage and
                               //and set Power-up Timer to 16msecs
_FGS(CODE_PROT_OFF);           //Set Code Protection Off for the
                               //General Segment
//----------------------------------------------------------------------

                        /*****************************
                        *          Cblage           *
                        *****************************/
/* Ports I2C
************
SDA sur port RF2 (broche 26)
SCL sur port RF3 (broche 25) */

/* Port srie asynchrone sur UART2
**********************************
RXD sur port RF4=U2RX (broche 28)
TXD sur port RF5=U2TX (broche 27) */

/* Signaux de test
******************
TEST0 sur port RF0 (broche 30) 
TEST1 sur port RF1 (broche 29) */
#define TEST0 LATFbits.LATF0 
#define TEST1 LATFbits.LATF1
#define TEST0_Dir TRISFbits.TRISF0
#define TEST1_Dir TRISFbits.TRISF1

                        /*****************************
                        *    Dfinitions de type     *
                        ******************************/
/* Informations de fonctionnement du controleur
***********************************************/
typedef struct
 {
  unsigned int Adr_I2C;    // Adresse du destinataire I2C
  unsigned int Adr_WR_E2P; // Adresse actuelle d'criture dans la mmoire E2
                           // Attention : il faut changer octets poids fort 
                           // et faible car transm. octets en adr. croissantes
  unsigned long NB_HALL; // Nb de chgts d'tats des capteurs Hall depuis
                         // la 1 mise en service
  struct tETATBAT
   {
    unsigned int VBAT;   // Tension batterie
    unsigned int IBAT;   // Courant batterie
    unsigned int QBAT;   // Quantit d'lectricit consomme
   } ETATBAT;
  unsigned long PERIOD_H;// Dure entre 2 chgts d'tats des capteurs Hall
  unsigned char PROFIL;  // Profil d'assistance
  unsigned char CLEF_OK; // Indication de prsence de la clef code
 } tDATASCONT;

typedef enum
{
  Start    = 0,
  MSend    = 1,
  MReceive = 2,
  Restart  = 3,
  MAck     = 4,
  Stop     = 5
}tSequenceI2C; // Type de squence I2C

typedef enum
{
  Write    = 0,
  Read     = 1
}tR_W; // Type "R/W" pour accs I2C

                        /*******************************
                        *   Constantes non mmorises  *
                        ********************************/
#define FCY 4000000*16/4  // xtal = 4Mhz; PLLx16 -> 16 MIPS

                        /*******************************
                        *  Constantes en mmoire flash *
                        ********************************/

                        /*******************************
                        *       Variables              *
                        ********************************/
/* Informations de fonctionnement du controleur
***********************************************/
tDATASCONT DATASCONT;
/* Liaison I2C
**************/
tSequenceI2C SequenceI2C_act; // Code de la squence I2C actuelle
tR_W R_W_I2C; // Type d'accs I2C : Write ou Read
unsigned int Nb_octets_I2C; // Nombre d'octets  transmettre  partir de 
                            // l'adresse pDatas_I2C 
unsigned int Cpt_octets_I2C;// Compteur d'octets transmis sur le bus I2C
unsigned char *pDatas_I2C;  // Utilis comme pointeur vers les donnes I2C
/* Buffers RX et TX pour l'UART
*******************************/
#define RX_BufSize 16         // Taille du Buffer_RX
char Buffer_RX[RX_BufSize];   // Buffer circulaire de rception
char *ptrRX_WRdata=Buffer_RX; // Pointeur d'criture ds Buffer_RX
char *ptrRX_RDdata=Buffer_RX; // Pointeur de lecture ds Buffer_RX
#define TX_BufSize 32         // Taille du Buffer_TX
char Buffer_TX[TX_BufSize];   // Buffer circulaire de transmission
char *ptrTX_WRdata=Buffer_TX; // Pointeur d'criture ds Buffer_TX
char *ptrTX_RDdata=Buffer_TX; // Pointeur de lecture ds Buffer_TX
/* Divers
*********/
char Compt_T1;

                        /*******************************
                        *          Fonctions           *
                        ********************************/

/*******************************
*           UART               *
********************************/

/*********************************************************************
  Function:        void InitUART(void)
  Description:     Initialisation de l'UART
                     - 38400 bauds, 8 bits, pas de parit, 1 bit Stop
                     - interruption sur TX si buffer vide
                     - interruption sur RX  chaque caractre
********************************************************************/
void InitUART(void)
{
 U2MODE=0x8000; // UART valid, 8 bits, pas de parit, 1 bit stop
 U2STA =0x8400; // Interruption TX si buffer vide
                // UART TX valid
                // Interruption RX  chaque caractre
 U2BRG =(FCY/(16*38400))-1; // 38400 bauds (erreur +0,16%)
}

/*********************************************************************
  Function:        int ReadRXD(void)
  Description:     Lecture d'un octet dans le buffer de rception
                     - renvoi de 0 si aucun caractre reu
                     - renvoi de 1 si caractre reu :
                        - caractre plac dans "c"
                        - ptrRX_RDdata incrment modulo RX_BufSize
********************************************************************/
int ReadRXD(char *c)
{
 if (ptrRX_RDdata==ptrRX_WRdata) return(0); // Pas de caractre reu
 else
 {
  *c=*ptrRX_RDdata++;
  if (ptrRX_RDdata==Buffer_RX+RX_BufSize) ptrRX_RDdata=Buffer_RX;
  return(1);
 }
}

/*********************************************************************
  Function:        void WriteTXD(char c)
  Description:     Ecriture d'un octet dans le buffer d'mission
                     - attendre libration U2TXREG
                     - envoi du caractre reu
                     - ptrTX_WRdata incrment modulo TX_BufSize
********************************************************************/
void WriteTXD(char c)
{
 int Nb_char;
 // Attendre libration du buffer
 do
  {
   // Calcul du nb de caractres restant  transmettre
   Nb_char=ptrTX_WRdata-ptrTX_RDdata;
   if (Nb_char < 0) Nb_char=Nb_char+TX_BufSize;
  } 
 while (Nb_char > (TX_BufSize-2));
 // Caractre plac dans le buffer circulaire
 *ptrTX_WRdata++=c; 
 if (ptrTX_WRdata==Buffer_TX+TX_BufSize) ptrTX_WRdata=Buffer_TX;
 // Provoquer la 1 interruption si 1 caractre
 if (Nb_char==0) IFS1bits.U2TXIF=1; 
}

/*********************************************************************
  Function:        void WriteIntTXD(int i)
  Description:     Ecriture d'un entier dans le buffer d'mission
********************************************************************/
void WriteIntTXD(int i)
{
 WriteTXD(i>>8); // Octet MSB d'abord
 WriteTXD(i);    // Puis octet LSB
}

/*******************************
*        Liaison I2C           *
********************************/

/*******************************************************************
  Function:    void InitI2C(void)
  Description: Initialisation du coupleur I2C :
                - mode Master
                - horloge SCL = 50kHz
                - interruptions "Master" valides
********************************************************************/
#define FSCL 50000 // Pour 50kHz
void InitI2C(void)
{
 I2CBRG=FCY/FSCL-FCY/1111111-1; // FSCL=50kHz
 I2CCON=0x8000; // I2CEN=0 : broches 26=SDA, broche 25=SCL
                // I2CSIDL=0 : module actif en mode "Idle"
                // SCLREL=0 : sans effet
                // IPMIEN=0 : mode IPMI inhib
                // A10M=0 : adresse I2C sur 7 bits
                // DISSLW=0 : slew rate control valid
                // SMEN=0 : seuils d'entre SMBus inhibs
                // GCEN=0 : pas d'appel gnral
                // STREN=0 : "clock stretching" inhib
                // ACKDT=0 : transmission ACK durant acquitement
                // ACKEN=RCEN=PEN=RSEN=SEN=0 : aucune squence I2C active
}

/*******************************************************************
  Function:    void Write_to_I2C(void)
  Description: criture de "Nb_octets_I2C" points par "pDatas_I2C" dans une
               EEPROM I2C
  Pr-requis : Les octets points par "pDatas_I2C" doivent tre affects.
               En particulier, le 1 est l'adresse I2C de la cible (dj
               dcale d'un cran  gauche).
               "Nb_octets_I2C" est affecte
  Traitement : - activation de la squence START
               - le programme d'interruption _MI2CInterrupt s'occupe des autres 
                 squences pour crire les octets ds l'E2P
********************************************************************/
void Write_to_I2C(void)
{
 R_W_I2C=Write;    // Ecriture ds l'E2P
 Cpt_octets_I2C=0; // Raz compteur d'octets (utilis par _MI2CInterrupt)
 SequenceI2C_act=Start; // Squence I2C actuelle
 I2CCONbits.SEN=1; // Activation de la squence START
                   // Une interruption est active par MI2CIF  la fin
                   // Le programme d'interruption _MI2CInterrupt se charge du
                   // reste du traitement
} 

/*******************************************************************
  Function:    void DATASCONT_to_I2C(void)
  Description: criture des champs NB_HALL, VBAT et IBAT de DATASCONT
               (8 octets) dans une EEPROM I2C,  partir de l'adresse "Adr"
  Pr-requis : Les champs Adr_I2C et Adr_WR_E2P de DATASCONT sont affects
  Traitement : - affectation de pDatas_I2C avec l'adresse de base de DATASCONT
               - affectation de Nb_octets_I2C avec 3+8 (3 octets : adresse I2C
                 et adresse E2P + 8 octets de datas)
               - activation de la squence START
               - le programme d'interruption _MI2CInterrupt s'occupe des autres 
                 squences pour crire les 8 octets  partir de pDatas_I2C ds l'E2P
********************************************************************/
void DATASCONT_to_I2C(void)
{
 pDatas_I2C=&DATASCONT;// Adresse de base des donnes  transmettre
                       // ("DATASCONT" est dj affecte)
 Nb_octets_I2C=8+3;    // Adresse I2C + adresse E2P + 8 octets de datas
 Write_to_I2C();       // Ecriture des Nb_octets_I2C points par pDatas_I2C
} 

/*******************************************************************
  Function:    void Read_I2C_to_RXD(void)
  Description: lecture de "Nb_octets_I2C" depuis l'EEPROM I2C et transfert
               vers TXD
  Pr-requis : Les premiers octets points par "pDatas_I2C" doivent tre 
               affects par l'adresse I2C de la cible (dj dcale d'un 
               cran  gauche) et l'adresse de base  lire ds l'EEPROM.
               La vaiable "Nb_octets_I2C" est affecte
  Traitement : - activation de la squence START
               - le programme d'interruption _MI2CInterrupt s'occupe des autres 
                 squences pour lire les Nb_octets_I2C ds l'E2P et les copier
                 dans le buffer TXD
********************************************************************/
void Read_I2C_to_RXD(void)
{
 R_W_I2C=Read;          // Lecture de l'E2P I2C
 SequenceI2C_act=Start; // Squence I2C actuelle
 Cpt_octets_I2C=0; // Raz compteur d'octets (utilis par _MI2CInterrupt)
 I2CCONbits.SEN=1; // Activation de la squence START
                   // Une interruption est active par MI2CIF  la fin
                   // Le programme d'interruption _MI2CInterrupt se charge du
                   // reste du traitement
} 

/*******************************************************************
  Function:    void DATASCONT_I2C_to_RXD(void)
  Description: lecture de Nb_octets_I2C depuis l'EEPROM I2C et transfert
               vers RXD avec l'entte "DATASCONT" en Ascii                
  Pr-requis : Les champs Adr_I2C et Adr_WR_E2P de DATASCONT sont affects
               La variable Nb_octets_I2C est affecte                  
  Traitement : - affectation de pDatas_I2C avec l'adresse de base de DATASCONT
               - affectation de Nb_octets_I2C avec 3+8 (3 octets : adresse I2C
                 et adresse E2P + 8 octets de datas)
               - activation de la squence START
               - le programme d'interruption _MI2CInterrupt s'occupe des autres 
                 squences pour lire les Nb_octets_I2C ds l'E2P et les copier
                 dans le buffer TXD
********************************************************************/
void DATASCONT_I2C_to_RXD(void)
{
 pDatas_I2C=&DATASCONT;// Adresse de base des donnes  transmettre
                       // ("DATASCONT" est affecte avec les adresses I2C et E2P)
 Read_I2C_to_RXD();    // Lecture de "Nb_octets_I2C" et transfert vers TXD
} 

                        /*******************************
                        * Fonctions d'initialisations  *
                        ********************************/

/*******************************************************************
  Function:        void InitTMR1(void)
  Description:     Initialisation du Timer 1 (Timer A) pour provoquer
                   une interruption  rythme rgulier.
  Note : le rythme est choisi pour obtenir une frquence des transmissions
         des paramtres sur I2C de 10Hz
********************************************************************/
void InitTMR1(void)
{
 T1CON=0x0030;	     // Fonctionnement continu, horloge=Fcy/256=2MHz
 TMR1=0;             // Raz du compteur pour dcaler la 1 interruption
 PR1=(FCY/256)/10-1; // Frquence = 10Hz
 T1CONbits.TON=1;    // Validation timer 1 
}

                       /*****************************
                       * Fonctions d'interruptions  *
                       ******************************/

/*******************************************************************
  Function:        void _ISR _T1Interrupt (void)
  Descrition :     Frquence = 10Hz
                   Traitement :  
                    - Transmission de Nb_octets_I2C  partir de pDatas sur I2C
********************************************************************/
void _ISR _T1Interrupt (void)
{
 IFS0bits.T1IF=0; // Acquitement interruption
 TEST0=!TEST0;    // Inversion TEST0 pour tests
 Nb_octets_I2C=8;
 //DATASCONT_I2C_to_RXD();
 Compt_T1++;
 DATASCONT_to_I2C();
}

/*******************************************************************
  Function:        void _ISR _MI2CInterrupt (void)
  Description :    Activ  chaque fin de squence I2C
                   Traitements :
                    - En criture :
                      - 1ire activation  la fin de la squence START 
                      - puis transmission des "Nb_octets_I2C"  partir de
                        pDatas_I2C sur le bus I2C
                      - activation squence STOP  la fin ou si NACK
                    - En lecture :
                      - 1ire activation  la fin de la squence START, puis
                      - transmission adresse I2C en criture,
                      - transmission adresse E2P  lire,
                      - activation RESTART,
                      - transmission adresse I2C en lecture,
                      - puis lecture de "Nb_octets_I2C" sur le bus I2C
                        et transfert sur TXD
                      - activation squence STOP  la fin ou si NACK
********************************************************************/
void _ISR _MI2CInterrupt (void)
{
 IFS0bits.MI2CIF=0; // Acquitement interruption
 if (R_W_I2C==Write)
 {
  // Accs en criture
  // *****************
  switch (SequenceI2C_act)
  {
   case Start : // Fin de la squence START ?
   {
    SequenceI2C_act=MSend;// Oui : dbut de la squence de transmission d'octets
    I2CTRN=*pDatas_I2C   ;// Dbut squence de transmission de l'adresse I2C en WR
                          // Prochaine interruption  la rception de l'acquitement
    pDatas_I2C++; // Pour pointer l'adresse E2P
    break;
   }
   case MSend : // Rception d'un acquitement de transmission d'octet ?
   {
    Cpt_octets_I2C++; // Incrmentation du compteur d'octets transmis
    pDatas_I2C++;
    if (I2CSTATbits.ACKSTAT) 
    {
     SequenceI2C_act=Stop; // NACK : arrt du transfert
     I2CCONbits.PEN=1;     // Activer la squence STOP
     break; 
    }
    if (Cpt_octets_I2C < Nb_octets_I2C) // Octet de donnes ?
    {
     I2CTRN=*pDatas_I2C;  // Oui : recherche de l'octet suivant   
     break;
    }
    if (Cpt_octets_I2C==Nb_octets_I2C) // Dernier octet transmis ?
    {
     SequenceI2C_act=Stop; // Arrt du transfert
     I2CCONbits.PEN=1;     // Activer la squence STOP
     break; 
    }
   }
   case Stop : // Fin de la squence STOP ?
   {
    break; // Oui : on ne fait rien !
   }
  }
 }
 else
 {
  // Accs en lecture
  // ****************
  switch (SequenceI2C_act)
  {
   case Start : // Fin de la squence START ?
   {
    I2CTRN=*pDatas_I2C; // Dbut squence de transmission de l'adresse I2C en WR
                        // Prochaine interruption  la rception de l'acquitement
    pDatas_I2C++;       // Pour pointer l'adresse E2P
    SequenceI2C_act=MSend;// Dbut de la squence de transm. de l'adresse
    break;
   }
   case MSend : // Fin de la transmission d'un octet ?
   {
    if (I2CSTATbits.ACKSTAT) // Transmission acquite ?
    {
     SequenceI2C_act=Stop; // NACK : arrt du transfert
     I2CCONbits.PEN=1;     // Activer la squence STOP
     break; 
    }
    Cpt_octets_I2C++; // Incrmentation du compteur d'octets transmis
    pDatas_I2C++;     // Pour pointer l'octet LSB de l'adresse E2P
    if (Cpt_octets_I2C < 3) // Transmission adresse E2P ?
    {
     I2CTRN=*pDatas_I2C;  // Oui : recherche de l'octet suivant   
     break;
    }
    if (Cpt_octets_I2C==3)  // Adresse complte transmise ?
    {
     SequenceI2C_act=Restart; // Oui : squence RESTART
     I2CCONbits.RSEN=1;       // Activer la squence RESTART
     break; 
    }
    if (Cpt_octets_I2C==4)  // Fin transmission adresse I2C aprs Restart ?
    {
     SequenceI2C_act=MReceive;// Oui :  dbut squence "rception"
     Cpt_octets_I2C=0;  // Pour compter les octets lus
     I2CCONbits.RCEN=1; // Activer la squence de rception d'un octet
     break; 
    }
   }
   case Restart : // Fin squence Restart ?
   {
    I2CCONbits.ACKDT=0; // Acquitement pour tous les octets lus, sauf le dernier
    I2CTRN=DATASCONT.Adr_I2C+1;// Aprs Restart : transmission adresse I2C en RD
    SequenceI2C_act=MSend;     // Prochaine interruption : test ACK
    break;
   }
   case MReceive : // Fin de rception d'un octet ?
   {
    Cpt_octets_I2C++;    // Incrmentation du compteur d'octets lus
    WriteTXD(I2CRCV);    // Octet reu transfr dans buffer TXD
    if (Cpt_octets_I2C==Nb_octets_I2C) I2CCONbits.ACKDT=1; // NACK pour le dernier octet
    I2CCONbits.ACKEN=1;  // Activer la squence ACK
    SequenceI2C_act=MAck;// Prochaine interruption  la fin de la squence
    break;
   }
   case MAck : // Fin d'une squence ACK ?
   {
    if (Cpt_octets_I2C==Nb_octets_I2C)
     {
      SequenceI2C_act=Stop; // Arrt du transfert  la lecture du dernier octet
      I2CCONbits.PEN=1;     // Activer la squence STOP
     }
    else 
     {
      SequenceI2C_act=MReceive;// Nouvelle squence "rception"
      I2CCONbits.RCEN=1; // Activer la squence de rception d'un octet
     }
    break;
   }
   case Stop : // Fin de la squence STOP ?
   {
    break; // Oui : on ne fait rien !
   }
  }
 }
}

/*********************************************************************
  Function:        void _ISR _U2TXInterrupt (void)
  Description :    Programme d'interruption dclench quand le buffer
                   de transmission de 4 octets est vide
********************************************************************/
void _ISR _U2TXInterrupt (void)
{
 IFS1bits.U2TXIF=0; // Clear interrupt flag
 if (ptrTX_RDdata!=ptrTX_WRdata)
 {
  U2TXREG=*ptrTX_RDdata++;
  if (ptrTX_RDdata==Buffer_TX+TX_BufSize) ptrTX_RDdata=Buffer_TX;
 }
}

/*********************************************************************
  Function:        void _ISR _U2RXInterrupt (void)
  Description :    Programme d'interruption dclench  la rception
                   d'un caractre
********************************************************************/
void _ISR _U2RXInterrupt (void)
{
 IFS1bits.U2RXIF=0; // Clear interrupt flag
 while (U2STAbits.URXDA)
 {
  *ptrRX_WRdata++=U2RXREG;
  if (ptrRX_WRdata==Buffer_RX+RX_BufSize) ptrRX_WRdata=Buffer_RX;
 }
}

                        /*****************************
                        *    Programme principal     *
                        ******************************/
int main (void)
{
 U2MODE=0; //Inhiber l'UART  cause du bootloader
 TEST0_Dir=0; // Signal TEST0 en sortie
 TEST1_Dir=0; // Signal TEST1 en sortie
 DATASCONT.Adr_I2C=0b10100000; // Adresse I2C d'un 24AA512 avec A2=A1=A0="0"
                               // (dj dcale  gauche d'un cran et en criture)
 DATASCONT.Adr_WR_E2P=0xAABB;  // Adresse d'criture dans l'EEPROM
 DATASCONT.NB_HALL=0x01020304;
 DATASCONT.ETATBAT.VBAT=0x0506;
 DATASCONT.ETATBAT.IBAT=0x0708;
 DATASCONT.ETATBAT.QBAT=0x090A;
 DATASCONT.PERIOD_H=0x0B0C0D0E;
 DATASCONT.PROFIL=0x0F;
 DATASCONT.CLEF_OK=0xFF;

 InitUART(); // 38400 bauds, 8 bits, pas de parit, 1 bit stop
 InitI2C();  // Initialisation du coupleur I2C: mode Master avec interruptions,
             // horloge SCL=10kHz
 InitTMR1(); // Initialisation Timer 1 pour RTI de 10Hz
 // Validations des interruptions
 IFS0bits.T1IF=0;   // Raz indicateur Timer 1 
 IEC0bits.T1IE=1;   // Validation interruption Timer 1
 IFS0bits.MI2CIF=0; // Raz indicateur MI2CIF
 IFS0bits.SI2CIF=0; // Raz indicateur SI2CIF
 IEC0bits.MI2CIE=1; // Validation interruptions "Master I2C"
 IEC0bits.SI2CIE=0; // Inhibition interruptions "Slave I2C"
 IFS1bits.U2TXIF=0; // Raz indicateur interruption TX sur UART2
 IEC1bits.U2TXIE=1; // Validation interruption TX sur UART2
 //IFS1bits.U2RXIF=0; // Raz indicateur interruption RX sur UART2
 //IEC1bits.U2RXIE=1; // Validation interruption RX sur UART2
 while (1) {};
}
